import logging
import os
import shutil
import tempfile
from io import StringIO
from pathlib import Path
from textwrap import dedent
from typing import List

import cwl_utils.parser_v1_1 as parser
import pytest
from cwltool.load_tool import fetch_document, resolve_and_validate_document
from dataclasses import dataclass
from WDL import Error, parse_document

from acclimatise import Command, WrapperGenerator
from acclimatise.name_generation import NameGenerationError
from acclimatise.yaml import yaml

logging.getLogger("cwltool").setLevel(30)


def skip_not_installed(executable):
    """
    Returns a pytest decorator to skip the test if the given executable is not in the path
    """
    return pytest.mark.skipif(
        not shutil.which(executable), reason="{} is not installed".format(executable)
    )


@dataclass
class HelpText:
    path: str
    cmd: List[str]
    positional: int
    named: int
    subcommands: int


all_tests = [
    pytest.param(
        HelpText(
            path="test_data/samtools_bedcov.txt",
            cmd=["samtools", "bedcov"],
            positional=2,
            named=4,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/gth.txt",
            cmd=["gth"],
            positional=0,
            named=37,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_closest.txt",
            cmd=["bedtools", "closest"],
            positional=0,
            named=29,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_spacing.txt",
            cmd=["bedtools", "spacing"],
            positional=0,
            named=5,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/samtools_quickcheck.txt",
            cmd=["samtools", "quickcheck"],
            positional=1,
            named=2,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_subtract.txt",
            cmd=["bedtools", "subtract"],
            positional=0,
            named=20,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_window.txt",
            cmd=["bedtools", "window"],
            positional=0,
            named=15,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/samtools_dict.txt",
            cmd=["samtools", "dict"],
            positional=1,
            named=5,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa_index.txt",
            cmd=["bwa", "index"],
            positional=1,
            named=4,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_multiinter.txt",
            cmd=["bedtools", "multiinter"],
            positional=0,
            named=8,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools_coverage.txt",
            cmd=["bedtools", "coverage"],
            positional=0,
            named=20,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa_mem.txt",
            cmd=["bwa", "mem"],
            positional=3,
            named=36,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa_bwt2sa.txt",
            cmd=["bwa", "bwt2sa"],
            named=1,
            positional=2,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa_bwtupdate.txt",
            cmd=["bwa", "bwtupdate"],
            named=0,
            positional=1,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/htseq_count.txt",
            cmd=["htseq-count"],
            named=20,
            positional=2,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bedtools.txt",
            cmd=["bedtools"],
            positional=0,
            named=1,
            subcommands=43,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/pisces.txt",
            cmd=["pisces"],
            named=57,
            positional=0,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/podchecker.txt",
            cmd=["podchecker"],
            named=2,
            positional=1,
            subcommands=0,
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/bwa.txt", cmd=["bwa"], named=0, positional=0, subcommands=14
        ),
    ),
    pytest.param(
        HelpText(
            path="test_data/samtools.txt",
            cmd=["samtools"],
            positional=0,
            named=0,
            subcommands=29,
        ),
    ),
    # These last two are really strange, maybe I'll support them eventually
    pytest.param(
        HelpText(
            path="test_data/dinosaur.txt",
            cmd=["dinosaur"],
            positional=0,
            named=24,
            subcommands=0,
        ),
        marks=pytest.mark.xfail,
    ),
    pytest.param(
        HelpText(
            path="test_data/mauve.txt",
            cmd=["mauveAligner"],
            positional=2,
            named=30,
            subcommands=0,
        ),
        marks=pytest.mark.xfail,
    ),
]

all_tests_lookup = {tuple(param.values[0].cmd): param.values[0] for param in all_tests}

all_ids = [" ".join(case.values[0].cmd) for case in all_tests]


def convert_validate(cmd: Command, lang: str = None, explore=True):
    """
    Converts the command into one or all formats, and then validates that they worked
    """
    if lang:
        conv = WrapperGenerator.choose_converter(lang)()
        with tempfile.TemporaryDirectory() as tempd:
            for path, subcommand in conv.generate_tree(cmd, Path(tempd)):
                content = path.read_text()
                validators[lang](content, subcommand, explore=explore)
    else:
        for lang in ("cwl", "wdl", "yml"):
            convert_validate(cmd, lang, explore=explore)


def validate_cwl(cwl: str, cmd: Command = None, explore: bool = True):
    parsed = yaml.load(cwl)
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        tmpfile = tmpdir / parsed["id"]
        tmpfile.write_text(cwl)
        loading_context, workflowobj, uri = fetch_document(str(tmpfile))
        resolve_and_validate_document(loading_context, workflowobj, uri)

        if cmd:
            assert len(workflowobj["inputs"]) == len(cmd.positional) + len(cmd.named)


def validate_wdl(wdl: str, cmd: Command = None, explore=True):
    """
    :param explore: If true, we're in explore mode, and we should ignore subcommands
    """
    try:
        doc = parse_document(wdl)
    except Error.SyntaxError as e:
        print(wdl)
        raise e

    task = doc.tasks[0]

    # task.inputs might be None
    cmd_input_names = {inp.name for inp in task.inputs or []}
    # Verify that every parameter has been documented
    wdl_input_names = {meta for meta in task.parameter_meta.keys()}
    assert cmd_input_names == wdl_input_names

    # wdl_input_descriptions = {meta for meta in doc.tasks[0].parameter_meta.values()}
    # cmd_input_descriptions = {arg.description for arg in [*cmd.subcommands, *cmd.named, *cmd.positional]}
    # assert wdl_input_descriptions == cmd_input_descriptions, wdl_input_descriptions ^ cmd_input_descriptions

    # Check that the generated WDL has the correct parameter meta fields
    if cmd:
        cmd_names = [arg.full_name() for arg in [*cmd.named, *cmd.positional]]
        target = len(cmd.named) + len(cmd.positional)
        # If we're not in explore mode, subcommands will be parsed as inputs
        if not explore:
            target += len(cmd.subcommands)
        assert len(task.parameter_meta) == target, "{} vs {}".format(
            task.parameter_meta.keys(), cmd_names
        )


def validate_yml(yml: str, cmd: Command = None, explore=True):
    stream = StringIO(yml)
    yaml.load(stream)


def ensure_conda():
    """
    Ensures we're in a conda env, and that the PATH is correctly set
    """
    if "CONDA_PREFIX" not in os.environ:
        raise Exception("Must be in a conda environment")
    bin = os.path.join(os.environ["CONDA_PREFIX"], "bin")
    os.environ["PATH"] = bin + os.pathsep + os.environ["PATH"]


validators = dict(cwl=validate_cwl, wdl=validate_wdl, yml=validate_yml)



# from acclimatise.flag_parser.elements import cli_id, any_flag, long_flag, short_flag, flag_with
from pathlib import Path

from acclimatise.flag_parser.elements import customIndentedBlock
from acclimatise.usage_parser.model import UsageElement

from .elements import *


def normalise_cline(tokens):
    """
    Normalise a command line string, such as ["dotnet", "Pisces.dll"], converting it to ["dotnet", "pisces"]
    :param tokens:
    :return:
    """
    return [Path(el.lower()).stem for el in tokens]


def parse_usage(cmd, text, debug=False):
    toks = usage.setDebug(debug).searchString(text)
    if not toks:
        # If we had no results, return an empty command
        return Command(command=cmd, positional=[], named=[])

    toks = toks[0]

    positional = [tok for tok in toks if isinstance(tok, UsageElement)]
    flags = [tok for tok in toks if isinstance(tok, Flag)]

    # Remove an "options" argument which is just a proxy for other flags
    # positional = [pos for pos in positional if pos.text.lower() != "options"]
    # The usage often starts with a re-iteration of the command name itself. Remove this if present
    for i in range(len(positional)):
        # For each positional argument, if the entire cmd string is present, slice away this and everything before it
        end = i + len(cmd)
        if end <= len(positional) and normalise_cline(
            [pos.text for pos in positional[i:end]]
        ) == normalise_cline(cmd):
            positional = positional[end:]

    if not any([tok for tok in positional if tok.variable]):
        # If the usage didn't explicitly mark anything as a variable using < > brackets, we have to assume that
        # everything other than flags are positional elements
        for element in positional:
            element.variable = True

    return Command(
        command=cmd,
        positional=[
            Positional(description="", position=i, name=el.text, optional=el.optional)
            for i, el in enumerate(positional)
        ],
        named=flags,
    )



"""
Tests for the name generation of single flags
"""
from acclimatise.converter.wdl import WdlGenerator
from acclimatise.model import EmptyFlagArg, Flag, Positional, SimpleFlagArg


def test_samtools_dict_output():
    gen = WdlGenerator()
    arg = Flag(
        synonyms=["-o", "--output"],
        description="file to write out dict file [stdout]",
        args=SimpleFlagArg(name="str"),
    )
    name = gen.choose_variable_names([arg])[0].name
    # The WDL converter should avoid naming a variable "output" since that's a WDL keyword
    assert name != "output"

    # Also, since we have a description, the generator shouldn't choose the lazy option of var_output
    assert name != "var_output"


def test_samtools_quickcheck_output():
    gen = WdlGenerator()
    arg = Positional(description="", position=0, name="input")
    name = gen.choose_variable_names([arg])[0].name
    # The WDL converter should avoid naming a variable "output" since that's a WDL keyword
    assert name != "input"


def test_bwt2sa_i(gen):
    arg = Flag(synonyms=["-i"], description="", args=SimpleFlagArg(name="32"))

    name = gen.choose_variable_names([arg])[0].name
    # 32 isn't a valid variable name, so the only option here is to use the letter i
    assert name == "i"


def test_name_to_words_symbol(gen):
    """
    Check that we can get an argument name even if the argument's flag is a symbol
    """
    arg = Flag(
        synonyms=["-@"],
        description="Number of additional threads to use",
        args=EmptyFlagArg(),
    )

    name = gen.choose_variable_names([arg])[0].name
    assert name == "at"


def test_name_to_words(gen):
    """
    Check that we can get an argument name even if the argument's flag is a symbol
    """
    arg = Flag(synonyms=["--genomepaths"], description="", args=EmptyFlagArg(),)

    name = gen.choose_variable_names([arg])[0].name
    assert "genome" in name
    assert "paths" in name
    # assert list(arg._name_from_name()) == ["genome", "paths"]


def test_bwa_mem_infq(gen):
    arg = Positional(name="in1.fq", description="", position=0)
    name = gen.choose_variable_names([arg])[0].name
    # name = arg.variable_name([])
    assert "1" in name or "one" in name
    assert "in" in name
    assert "fq" in name



from .util import convert_validate


def test_premade_samtools(samtools_cmd):
    """
    Use a command tree that was generated beforehand, to quickly detect issues relating to the conversion of command
    trees
    """
    convert_validate(samtools_cmd, explore=True)


def test_premade_bedtools(bedtools_cmd):
    """
    Use a command tree that was generated beforehand, to quickly detect issues relating to the conversion of command
    trees
    """
    convert_validate(bedtools_cmd, explore=True)



import spacy
import wordsegment

# We load the spacy and the wordsegment models here as a kind of singleton pattern, to avoid multiple functions loading
# redundant copies

if len(wordsegment.WORDS) == 0:
    wordsegment.load()


try:
    nlp = spacy.load("en")
except IOError:
    raise Exception(
        "Spacy model doesn't exist! Install it with `python -m spacy download en`"
    )



"""
Functions that relate to executing the programs of interest, in order to extract their help text
"""
import os
import pty
import signal
import subprocess
import sys
import typing

import psutil


def kill_proc_tree(
    pid, sig=signal.SIGTERM, include_parent=True, timeout=None, on_terminate=None
):
    """
    Kill a process tree (including grandchildren) with signal "sig" and return a (gone, still_alive) tuple.
    "on_terminate", if specified, is a callabck function which is called as soon as a child terminates.

    Taken from https://psutil.readthedocs.io/en/latest/#kill-process-tree
    """
    assert pid != os.getpid(), "won't kill myself"
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    if include_parent:
        children.append(parent)
    for p in children:
        p.send_signal(sig)


def execute_cmd(help_cmd: typing.List[str], timeout: int = 5, **kwargs) -> str:
    """
    Execute a command defined by a list of arguments, and return the result as a string
    """
    master, slave = pty.openpty()
    popen_kwargs = dict(
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=slave, encoding="utf-8"
    )
    popen_kwargs.update(kwargs)

    # This works a lot like subprocess.run, but we need access to the pid in order to kill the process tree, so use Popen
    with subprocess.Popen(help_cmd, **popen_kwargs) as process:
        try:
            stdout, stderr = process.communicate(timeout=timeout)
        except subprocess.TimeoutExpired as e:
            # Kill the entire process tree, because sometimes killing the parent isn't enough
            kill_proc_tree(
                process.pid,
                include_parent=True,
                sig=signal.SIGKILL if sys.platform == "linux" else None,
            )
            process.communicate()
            return ""
        finally:
            os.close(master)
            os.close(slave)

    return stdout or stderr



import pytest

from acclimatise.converter import WrapperGenerator


@pytest.fixture()
def snake_gen():
    return WrapperGenerator(case="snake", generate_names=True)


@pytest.fixture()
def camel_gen():
    return WrapperGenerator(case="camel", generate_names=True)


@pytest.fixture()
def gen():
    return WrapperGenerator()



"""
Contains the CLI data model
"""
import abc
import enum
import itertools
import re
import typing
import unicodedata
from abc import abstractmethod

import spacy
from dataclasses import InitVar, dataclass, field
from ruamel.yaml import YAML, yaml_object
from spacy import tokens
from word2number import w2n

from acclimatise import cli_types
from acclimatise.name_generation import generate_name, segment_string
from acclimatise.nlp import wordsegment
from acclimatise.yaml import yaml


def useless_name(name: typing.List[str]):
    """
    Returns true if this name (sequence of strings) shouldn't be used as a variable name because it's too short and
    uninformative. This includes an entirely numeric name, which is almost never what you want
    """
    joined = "".join(name)
    if len(name) < 1 or len(joined) <= 1 or joined.isnumeric():
        return True

    # Numeric names are not useful
    try:
        if all([w2n.word_to_num(tok) is not None for tok in name]):
            return True
    except Exception:
        pass

    return False


@yaml_object(yaml)
@dataclass
class Command:
    """
    Class representing an entire command or subcommand, e.g. `bwa mem` or `grep`
    """

    def __post_init__(self):
        # Store certain special flags in their own fields
        if self.help_flag is None:
            for flag in self.named:
                if (
                    "--help" in flag.synonyms
                    or "-help" in flag.synonyms
                    or ("-h" in flag.synonyms and isinstance(flag.args, EmptyFlagArg))
                ):
                    self.help_flag = flag
                    self.named.remove(flag)

        if self.version_flag is None:
            for flag in self.named:
                if "--version" in flag.synonyms:
                    self.version_flag = flag
                    self.named.remove(flag)

        if self.usage_flag is None:
            for flag in self.named:
                if "--usage" in flag.synonyms:
                    self.usage_flag = flag
                    self.named.remove(flag)

    @property
    def as_filename(self) -> str:
        """
        Returns a sample filename that might be used to store this command (without a suffix)
        """
        return "_".join(self.command).replace("-", "_")

    @property
    def depth(self) -> int:
        """
        Returns the "depth" of this command, aka how many ancestors it has. An orphan command has depth 0, a subcommand
        has depth 1, a sub-sub-command has depth 2 etc
        """
        cmd = self
        depth = 0
        while cmd.parent is not None:
            cmd = cmd.parent
            depth += 1

        return depth

    def command_tree(self) -> typing.Generator["Command", None, None]:
        """
        Returns a generator over the entire command tree. e.g. if this command has 2 subcommands, each with 2
            subcommands, this will return a generator with 7 Commands
        """
        yield self
        for command in self.subcommands:
            yield from command.command_tree()

    positional: typing.List["Positional"]
    """
    All positional arguments supported by this command
    """

    named: typing.List["Flag"]
    """
    All named arguments (flags) supported by this command
    """

    command: typing.List[str]
    """
    The command line used to invoke this command, e.g. ["bwa", "mem"]
    """

    parent: typing.Optional["Command"] = None
    """
    The parent command, if this is a subcommand
    """

    subcommands: typing.List["Command"] = field(default_factory=list)
    """
    A list of subcommands of this command, e.g. "bwa" has the subcommand "bwa mem"
    """

    help_flag: typing.Optional["Flag"] = None
    """
    If identified, this is the flag that returns help text
    """

    usage_flag: typing.Optional["Flag"] = None
    """
    If identified, this is the flag that returns usage examples
    """

    version_flag: typing.Optional["Flag"] = None
    """
    If identified, this is the flag that returns the version of the executable
    """

    help_text: typing.Optional[str] = None
    """
    Optionally, the entire help text that was used to generate this Command
    """

    generated_using: typing.Optional[str] = None
    """
    Optionally, the flag that was used to generate this command. Often this will be the same as the help_flag
    """


@yaml_object(yaml)
@dataclass(unsafe_hash=True)
class CliArgument:
    """
    A generic parent class for both named and positional CLI arguments
    """

    description: str
    """
    Description of the function of this argument
    """

    @abstractmethod
    def argument_name(self) -> typing.List[str]:
        return []

    @abstractmethod
    def full_name(self) -> str:
        """
        Return a human-readable representation of this argument
        """
        pass

    @abstractmethod
    def get_type(self) -> cli_types.CliType:
        """
        Return a type object indicating the type of data this argument holds. e.g. If it's an array type this will be a
        CliList.
        """
        pass


@yaml_object(yaml)
@dataclass
class Positional(CliArgument):
    """
    A positional command-line argument. This probably means that it is required, and has no arguments like flags do
    """

    def full_name(self) -> str:
        """
        Getting the full name for a positional argument is easy - it's just the parameter name
        """
        return self.name

    position: int
    """
    The position in the command line that this argument must occupy
    """

    name: str
    """
    The name of this argument
    """

    description: str
    """
    A description of the function of this argument
    """

    optional: bool = False
    """
    If true, this argument is not required
    """

    def get_type(self) -> cli_types.CliType:
        # Try the the flag name, then the description in that order

        name_type = infer_type(self.name)
        if name_type is not None:
            return name_type

        flag_type = infer_type(self.full_name())
        if flag_type is not None:
            return flag_type

        return cli_types.CliString()


@yaml_object(yaml)
@dataclass(unsafe_hash=True)
class Flag(CliArgument):
    """
    Represents one single flag, with all synonyms for it, and all arguments, e.g. `-h, --help`
    """

    synonyms: typing.List[str]
    """
    A list of different ways to invoke this same option, e.g. ``-v`` and ``--verbose``
    """

    description: typing.Optional[str]
    """
    A description of the function of this flag
    """

    args: "FlagArg"
    """
    Describes the arguments to this flag, e.g. ``-n 1`` has a single numeric argument
    """

    optional: bool = True
    """
    If true, this flag is not required (the default)
    """

    def argument_text(self) -> typing.List[str]:
        return self.args.text()

    def variable_name(
        self, description_name: typing.List[str] = []
    ) -> typing.List[str]:
        """
        Returns a list of words that should be used in a variable name for this argument
        """
        # The super method returns the best name from the flag name or the description. If neither is sufficient, use
        # the argument to generate a name
        best = super().variable_name(description_name)

        if useless_name(best):
            nfa = list(self._name_from_arg)
            if not useless_name(nfa):
                return nfa

        return best

    @property
    def _name_from_arg(self) -> typing.Iterable[str]:
        """
        Generate a 1-3 word variable name for this flag, by parsing the description
        """
        if self.args is not None and hasattr(self.args, "name"):
            return segment_string(self.args.name)
        else:
            return []

    def get_type(self) -> cli_types.CliType:
        # Try the argument name, then the flag name, then the description in that order
        arg_type = self.args.get_type()
        if arg_type is not None:
            return arg_type

        flag_type = infer_type(self.full_name())
        if flag_type is not None:
            return flag_type

        description_type = infer_type(self.description)
        if description_type is not None:
            return description_type

        return cli_types.CliString()

    def full_name(self) -> str:
        """
        Getting the full name for a named flag is slightly harder, we need to find the longest synonym
        """
        return self.longest_synonym

    @staticmethod
    def from_synonyms(
        synonyms: typing.Iterable["FlagSynonym"], description: typing.Optional[str]
    ):
        """
        Creates a usable Flag object by combining the synonyms provided
        """
        synonym_str = []
        args = None
        arg_count = float("-inf")

        for synonym in synonyms:
            synonym_str.append(synonym.name)
            if synonym.argtype.num_args() > arg_count:
                arg_count = synonym.argtype.num_args()
                args = synonym.argtype

        return Flag(synonyms=synonym_str, args=args, description=description)

    @property
    def longest_synonym(self) -> str:
        """
        Returns the longest synonym this flag has. e.g. for `-h, --help`, it will return `--help`
        """
        return max(self.synonyms, key=lambda synonym: len(synonym))

    @property
    def shortest_synonym(self) -> str:
        """
        Returns the shortest synonym this flag has. e.g. for `-h, --help`, it will return `-h`
        """
        return min(self.synonyms, key=lambda synonym: len(synonym))


@yaml_object(yaml)
@dataclass
class FlagSynonym:
    """
    Internal class for storing the arguments for a single synonym
    """

    name: str
    """
    The entire flag string, e.g. "-n" or "--lines"
    """

    argtype: "FlagArg"
    """
    The number and type of arguments that this flag takes
    """

    @property
    def capital(self):
        return "".join(
            [
                segment.capitalize()
                for segment in re.split("[-_]", self.name.lstrip("-"))
            ]
        )


int_re = re.compile("(int(eger)?)|size|length|max|min", flags=re.IGNORECASE)
str_re = re.compile("str(ing)?", flags=re.IGNORECASE)
float_re = re.compile("float|decimal", flags=re.IGNORECASE)
bool_re = re.compile("bool(ean)?", flags=re.IGNORECASE)
file_re = re.compile("file|path", flags=re.IGNORECASE)
dir_re = re.compile("folder|directory", flags=re.IGNORECASE)


def infer_type(string) -> typing.Optional[cli_types.CliType]:
    """
    Reads a string (argument description etc) to find hints about what type this argument might be. This is
    generally called by the get_type() methods
    """
    if bool_re.match(string):
        return cli_types.CliBoolean()
    elif float_re.match(string):
        return cli_types.CliFloat()
    elif int_re.match(string):
        return cli_types.CliInteger()
    elif file_re.match(string):
        return cli_types.CliFile()
    elif dir_re.match(string):
        return cli_types.CliDir()
    elif str_re.match(string):
        return cli_types.CliString()
    else:
        return cli_types.CliString()


@yaml_object(yaml)
@dataclass
class FlagArg(abc.ABC):
    """
    The data model for the argument or arguments for a flag, for example a flag might have no arguments, it might have
    one argument, it might accept one option from a list of options, or it might accept an arbitrary number of inputs
    """

    def text(self) -> typing.List[str]:
        """
        Returns the text of the argument, e.g. for name generation purposes
        """
        return []

    @abc.abstractmethod
    def get_type(self) -> cli_types.CliType:
        """
        Return a type object indicating the type of data this argument holds. e.g. If it's an array type this will be a
        CliList.
        """
        pass

    @abc.abstractmethod
    def num_args(self) -> int:
        """
        Calculate the multiplicity of this argument
        """
        pass


@yaml_object(yaml)
@dataclass
class EmptyFlagArg(FlagArg):
    """
    A flag that has no arguments, e.g. `--quiet` that is either present or not present
    """

    def num_args(self) -> int:
        return 0

    def get_type(self):
        return cli_types.CliBoolean()


@yaml_object(yaml)
@dataclass
class OptionalFlagArg(FlagArg):
    """
    When the flag has multiple arguments, some of which are optional, e.g.
    -I FLOAT[,FLOAT[,INT[,INT]]]
    """

    names: list
    """
    Names of each argument
    """

    separator: str
    """
    Separator between each argument
    """

    def text(self) -> typing.List[str]:
        return list(
            itertools.chain.from_iterable(
                [wordsegment.segment(name) for name in self.names]
            )
        )

    def num_args(self) -> int:
        return len(self.names)

    def get_type(self):
        return cli_types.CliTuple([infer_type(arg) for arg in self.names])


@yaml_object(yaml)
@dataclass
class SimpleFlagArg(FlagArg):
    """
    When a flag has one single argument, e.g. `-e PATTERN`, where PATTERN is the argument
    """

    name: str
    """
    Name of this argument
    """

    def text(self) -> typing.List[str]:
        return list(wordsegment.segment(self.name))

    def num_args(self) -> int:
        return 1

    def get_type(self):
        return infer_type(self.name)


@yaml_object(yaml)
@dataclass
class RepeatFlagArg(FlagArg):
    """
    When a flag accepts 1 or more arguments, e.g. `--samout SAMOUTS [SAMOUTS ...]`
    """

    name: str
    """
    The name of this argument
    """

    def text(self) -> typing.List[str]:
        return list(wordsegment.segment(self.name))

    def num_args(self) -> int:
        return 1

    def get_type(self):
        t = infer_type(self.name)
        return cli_types.CliList(t)


@yaml_object(yaml)
@dataclass
class ChoiceFlagArg(FlagArg):
    """
    When a flag accepts one option from a list of options, e.g. `-s {yes,no,reverse}`
    """

    choices: typing.Set[str]
    """
    Set of possible choices that could be used for this argument
    """

    def text(self) -> typing.List[str]:
        return list(
            itertools.chain.from_iterable(
                [wordsegment.segment(name) for name in self.choices]
            )
        )

    def get_type(self):
        e = enum.Enum(
            "".join([choice.capitalize() for choice in self.choices]),
            list(self.choices),
        )
        return cli_types.CliEnum(e)

    def num_args(self) -> int:
        return 1



import inspect
import tempfile
from io import IOBase, StringIO, TextIOBase
from os import PathLike
from pathlib import Path
from typing import Generator, List

from cwl_utils.parser_v1_1 import (
    CommandInputParameter,
    CommandLineBinding,
    CommandLineTool,
)
from dataclasses import dataclass

from acclimatise import cli_types
from acclimatise.converter import NamedArgument, WrapperGenerator
from acclimatise.model import CliArgument, Command, Flag, Positional
from acclimatise.yaml import yaml

# def with_default_none(func, *args, **kwargs):
#     """
#     Calls a function, and inserts None for all arguments you didn't provide
#     """
#     spec = inspect.getfullargspec(func)
#     defaults = {arg: None for arg in spec.args}
#     defaults.pop("self", None)
#     return func(*args, **{**defaults, **kwargs})


@dataclass
class CwlGenerator(WrapperGenerator):
    case = "snake"

    @classmethod
    def format(cls) -> str:
        return "cwl"

    @staticmethod
    def snake_case(words: list):
        return "_".join([word.lower() for word in words])

    @staticmethod
    def to_cwl_type(typ: cli_types.CliType):
        if isinstance(typ, cli_types.CliFile):
            return "File"
        elif isinstance(typ, cli_types.CliDir):
            return "Directory"
        elif isinstance(typ, cli_types.CliString):
            return "string"
        elif isinstance(typ, cli_types.CliFloat):
            return "double"
        elif isinstance(typ, cli_types.CliInteger):
            return "long"
        elif isinstance(typ, cli_types.CliBoolean):
            return "boolean"
        elif isinstance(typ, cli_types.CliEnum):
            return "string"
        elif isinstance(typ, cli_types.CliList):
            return CwlGenerator.to_cwl_type(typ.value) + "[]"
        elif isinstance(typ, cli_types.CliTuple):
            return [CwlGenerator.to_cwl_type(subtype) for subtype in set(typ.values)]
        else:
            raise Exception(f"Invalid type {typ}!")

    def command_to_tool(self, cmd: Command) -> CommandLineTool:
        """
        Outputs the CWL wrapper to the provided file
        """
        inputs: List[CliArgument] = [*cmd.named] + (
            [] if self.ignore_positionals else [*cmd.positional]
        )
        names = self.choose_variable_names(inputs)

        tool = CommandLineTool(
            id=cmd.as_filename + ".cwl",
            baseCommand=list(cmd.command),
            cwlVersion="v1.1",
            inputs=[],
            outputs=[],
        )

        for arg in names:
            assert arg.name != "", arg
            tool.inputs.append(
                CommandInputParameter(
                    id=arg.name,
                    type=self.to_cwl_type(arg.arg.get_type()),
                    inputBinding=CommandLineBinding(
                        position=arg.arg.position
                        if isinstance(arg.arg, Positional)
                        else None,
                        prefix=arg.arg.longest_synonym
                        if isinstance(arg.arg, Flag)
                        else None,
                    ),
                    doc=arg.arg.description,
                )
            )
        return tool

    @property
    def suffix(self) -> str:
        return ".cwl"

    def save_to_string(self, cmd: Command) -> str:
        io = StringIO()
        yaml.dump(self.command_to_tool(cmd).save(), io)
        return io.getvalue()

    def save_to_file(self, cmd: Command, path: Path) -> None:
        map = self.command_to_tool(cmd).save(base_url=str(path))
        with path.open("w") as fp:
            yaml.dump(map, fp)



import tempfile
from itertools import product

import pytest
from pkg_resources import resource_filename

from acclimatise import parse_help

from .util import (
    HelpText,
    all_ids,
    all_tests,
    all_tests_lookup,
    convert_validate,
    validate_cwl,
    validate_wdl,
)


@pytest.mark.parametrize("test", all_tests, ids=all_ids)
def test_all(test: HelpText):
    """
    A comprehensive end-to-end test that tests the parser and converters, using the test data files
    """
    with open(resource_filename("test", test.path)) as fp:
        help_text = fp.read()

    cmd = parse_help(test.cmd, help_text)

    # Check we parsed the arguments correctly
    # Since we aren't exploring here, we can't distinguish between positionals and subcommands, so we can sum them
    assert len(cmd.positional) == test.positional + test.subcommands
    assert len(cmd.named) == test.named

    # Check that the help text is included in the command
    assert cmd.help_text == help_text

    # Check the converters work
    convert_validate(cmd, explore=False)



import logging
import pty
import subprocess
import typing

import dataclasses
import psutil
from pyparsing import ParseBaseException

from acclimatise.converter import WrapperGenerator
from acclimatise.converter.cwl import CwlGenerator
from acclimatise.converter.wdl import WdlGenerator
from acclimatise.converter.yml import YmlGenerator
from acclimatise.execution import execute_cmd
from acclimatise.flag_parser.parser import CliParser
from acclimatise.model import Command, Flag
from acclimatise.usage_parser import parse_usage

logger = logging.getLogger("acclimatise")


def _combine_flags(
    flag_lists: typing.Iterable[typing.Iterable[Flag]],
) -> typing.Iterable[Flag]:
    """
    Combines the flags from several sources, choosing the first one preferentially
    """
    lookup = {}

    # Build a list of flags, but only ever choose the first instance of a synonym
    for flags in flag_lists:
        for flag in flags:
            for synonym in flag.synonyms:
                stripped = synonym.lstrip("-")
                if stripped not in lookup:
                    lookup[stripped] = flag

    # Now, make them unique by description
    unique = {flag.longest_synonym: flag for flag in lookup.values()}

    return list(unique.values())


def parse_help(
    cmd: typing.Collection[str], text: str, parse_positionals=True
) -> Command:
    """
    Parse a string of help text into a Command. Use this if you already have run the executable and extracted the
    help text yourself

    :param cmd: List of arguments used to generate this help text, e.g. ['bwa', 'mem']
    :param text: The help text to parse
    :param parse_positionals: If false, don't parse positional arguments
    """
    help_command = CliParser(parse_positionals=parse_positionals).parse_command(
        name=cmd, cmd=text
    )
    usage_command = parse_usage(cmd, text)

    # Combine the two commands by picking from the help_command where possible, otherwise falling back on the usage
    fields = dict(
        help_text=text,
        # Use the help command's positionals preferentially, but fall back to usage
        positional=help_command.positional or usage_command.positional,
        # Combine the flags from both help and usage
        named=list(_combine_flags([help_command.named, usage_command.named])),
    )
    for field in dataclasses.fields(Command):
        fields[field.name] = (
            fields.get(field.name)
            or getattr(help_command, field.name)
            or getattr(usage_command, field.name)
        )

    return Command(**fields)


def best_cmd(
    cmd: typing.List[str],
    flags: typing.Iterable[str] = (["--help"], ["-h"], [], ["--usage"]),
    run_kwargs: dict = {},
) -> Command:
    """
    Determine the best Command instance for a given command line tool, by trying many
    different help flags, such as --help and -h, then return the Command. Use this if you know the command you want to
    parse, but you don't know which flags it responds to with help text. Unlike :py:func:`aclimatise.explore_command`,
    this doesn't even attempt to parse subcommands.

    :param cmd: The command to analyse, e.g. ['wc'] or ['bwa', 'mem']
    :param flags: A list of help flags to try, e.g. ['--help', '-h'], in order how which one you would prefer to use.
    Generally [] aka no flags should be last
    :param run_kwargs: kwargs to pass into subprocess.run, when we run the executable
    """
    # For each help flag, run the command and then try to parse it
    logger.info("Trying flags for {}".format(" ".join(cmd)))
    commands = []
    for flag in flags:
        help_cmd = cmd + flag
        logger.info("Trying {}".format(" ".join(help_cmd)))
        try:
            final = execute_cmd(help_cmd, **run_kwargs)
            result = parse_help(cmd, final)
            result.generated_using = flag
            commands.append(result)
        except (ParseBaseException, UnicodeDecodeError) as e:
            # If parsing fails, this wasn't the right flag to use
            continue

    # Sort by flags primarily, and if they're equal, return the command with the longest help text, and if they're equal
    # return the command with the most help flags. This helps ensure we get ["bedtools", "--help"] instead of
    # ["bedtools"]
    best = max(
        commands,
        key=lambda com: (
            len(com.named) + len(com.positional),
            # len(com.help_text) if com.help_text else 0,
        ),
    )
    logger.info(
        "The best help flag seems to be {}".format(
            " ".join(best.command + best.generated_using)
        )
    )
    return best


def is_subcommand(command: Command, parent: Command) -> bool:
    """
    Returns true if command is a valid subcommand, relative to its parent
    """
    # Recursively call this on all ancestors
    if parent.parent is not None and not is_subcommand(command, parent.parent):
        return False

    # This isn't a subcommand if it has the same input text as the parent
    if command.help_text and command.help_text == parent.help_text:
        return False

    # This isn't a subcommand if it has no flags
    if len(command.positional) + len(command.named) == 0:
        return False

    # This isn't a subcommand if it shares any positional with the parent command
    for pos_a, pos_b in zip(parent.positional, command.positional):
        if pos_a == pos_b:
            return False

    # This isn't a subcommand if it shares any flags with the parent command
    for flag_a, flag_b in zip(parent.named, command.named):
        if flag_a == flag_b:
            return False

    return True


def explore_command(
    cmd: typing.List[str],
    flags: typing.Iterable[str] = (["--help"], ["-h"], [], ["--usage"]),
    parent: typing.Optional[Command] = None,
    run_kwargs: dict = {},
    max_depth: int = 2,
    try_subcommand_flags=True,
) -> typing.Optional[Command]:
    """
    Given a command to start with, builds a model of this command and all its subcommands (if they exist).
    Use this if you know the command you want to parse, you don't know which flags it responds to with help text, and
    you want to include subcommands.

    :param cmd: Command line executable and arguments to explore
    :param flags: A list of help flags to try, e.g. ['--help', '-h'], in order how which one you would prefer to use.
    Generally [] aka no flags should be last
    :param parent: A parent Command to add this command to as a subcommand, if this command actually exists
    :param run_kwargs: kwargs to pass into subprocess.run, when we run the executable
    :param try_subcommand_flags: If true, try all the ``flags`` on each subcommand. If False, we choose
    the best help flag on the parent command and then use that same one on each child. Generally True is recommended
    since some tools (such as bedtools) use different help flags for subcommands
    """
    logger.info("Exploring {}".format(" ".join(cmd)))
    command = best_cmd(cmd, flags, run_kwargs=run_kwargs)

    # Check if this is a valid subcommand
    if parent:
        if is_subcommand(command, parent):
            logger.info("{} seems to be a valid subcommand".format(" ".join(cmd)))
            command.parent = parent
        else:
            logger.info(
                "{} does not seem to be a valid subcommand".format(" ".join(cmd))
            )
            return None

    # Recursively call this function on positionals, but only do this if we aren't at max depth
    if command.depth < max_depth:
        # By default we use the best parent help-flag
        child_flags = flags if try_subcommand_flags else [command.generated_using]
        for positional in command.positional:
            subcommand = explore_command(
                cmd=cmd + [positional.name],
                flags=child_flags,
                parent=command,
                run_kwargs=run_kwargs,
                max_depth=max_depth,
                try_subcommand_flags=try_subcommand_flags,
            )
            if subcommand is not None:
                command.subcommands.append(subcommand)

                # If we had any subcommands then we probably don't have any positionals, or at least don't care about them
                command.positional = []

    return command
